/* ---------------------------------------------------------- 
%   (C)1994,1995 Institute for New Generation Computer Technology 
%       (Read COPYRIGHT for detailed information.) 
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#ifdef DIST

#include <stdio.h>

#include <klic/basic.h>  /* fatal, klic_fprintf */
#include <klic/struct.h>
#include <klic/primitives.h>
#include <klic/unify.h>
#include <klic/index.h>
#include <klic/gb.h>
#include <klic/bb.h>
#include <klic/g_basic.h>
#include <klic/g_extern.h>
#include <klic/susp.h>
#include <klic/gg_macro.h>

#include "interpe.h"
#include <klic/distio.h>

#include "atom.h"
#include "funct.h"
#include "ge_exref.h"

#define CLASS_NAME  exref
#define G_OBJ_TYPE  struct exref_object

#include <klic/gc_macro.h>
#include <klic/gd_macro.h>
#include <klic/generic.h>

extern q read_hook_g_new();
extern void decode_exref();

GGDEF_GENERATE()
{
  G_STD_DECL;
  q rdhok;
  q* allocp = heapp();

  generic_arg[0] = makeint(g_self->node);
  generic_arg[1] = makeint(g_self->index);
  generic_arg[2] = makeint(g_self->wec);

  new_generic(read_hook_g_new, 3, rdhok);

  g_self->method_table = 0;

  g_self->node = 0;
  g_self->index = 0;
  g_self->wec = 0;

  g_self->gc_flag = IMPREC_RELEASED;

  return rdhok;
}

GGDEF_UNIFY()
{
  G_STD_DECL;
  struct generator_susp* gsusp = generator_suspp(derefone(g_self));
  struct exref_object* GG_OBJ =
    (struct exref_object*) untag_generator_susp(gsusp->u.o);

  G_SWITCH_ON_TERM(list0, atomic0, functor0, generic_data0, susp0);

 list0:
 atomic0:
 functor0:
 generic_data0:
  send_unify(GG_OBJ->node, GG_OBJ->index, GG_OBJ->wec, g_term);

  derefone(g_self) = g_term;

  GG_OBJ->method_table = NULL;
  GG_OBJ->gc_flag = IMPREC_RELEASED;
  GG_TERMINATE;

 susp0:
  {
    struct susprec* gterm = (struct susprec*) derefone(g_term);

    if( is_generator_susp(gterm->u) ){
      struct generator_susp* gsusp1 = generator_suspp(derefone(g_term));
      struct exref_object* dummy_obj =
	(struct exref_object*) untag_generator_susp(gsusp1->u.o);
      /* fprintf(stderr, "Node:%d unify EXREF-GENERA", my_node); */

      if( dummy_obj->method_table == get_exref_methtab() ){
	/* fprintf(stderr, "Node:%d unify EXREF-EXREF", my_node); */
      }else{
	fatal("unify EXREF-UnknownGEN occur");
      }

      goto list0;
    }else{
      /* fprintf(stderr, "Node:%d unify EXREF-non-EXREF", my_node); */
    }
  }
  return 0;  /* Cannot unify now */
}

G_DEF_ENCODE()
{
  long send_wec;

  if( g_self->wec < MIN_WEC ){
    fatal("invalid WEC in exref object");
  }

  if( g_self->wec < MIN_WEC*2 ){
    return GENERIC_FAILED;
  }
  send_wec = (g_self->wec) >> 1;

  g_self->wec -= send_wec;

  PUT_BUFFER(buffer, decode_exref);
  PUT_BUFFER(buffer, g_self->node);
  PUT_BUFFER(buffer, g_self->index);
  PUT_BUFFER(buffer, send_wec);

  return GENERIC_SUCCEEDED;
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  INT_CL_DEBUG_X(klic_fprintf(stdout, "exref GC invoked\n"));

  if( g_self->gc_flag != IMPREC_NOT_COPIED ){
    ioeprintf("invalid exref in GC\n");
    ERROR_STOP;
  }

  G_SET_NEWOBJ_IN_NEWGEN(newself);

  newself->node = g_self->node;
  newself->index = g_self->index;
  newself->wec = g_self->wec;
  G_COPY_KL1_TERM_TO_NEWGEN(g_self->to_exref, newself->to_exref);

  g_self->gc_flag = IMPREC_COPIED;
  newself->gc_flag = IMPREC_NOT_COPIED;

  G_RETURN_FROM_GC(newself);
}

#define GGUSE_MY_GENERATE
#define GGUSE_MY_UNIFY
#define GGUSE_MY_GC
#define GGUSE_MY_ENCODE

#include <klic/gg_methtab.h>

G_DEF_NEW() /* node, index, wec */
{
  G_STD_DECL_FOR_NEW();
  G_OBJ_TYPE* new_exref;

  G_SET_NEWOBJ_FOR_NEW(new_exref, G_OBJ_SIZE);

  new_exref->node  = (long) intval(g_argv[0]);
  new_exref->index = (long) intval(g_argv[1]);
  new_exref->wec   = (long) intval(g_argv[2]);

  new_exref->to_exref = GG_MAKE_HOOK_VAR((struct generator_object*) new_exref);

  new_exref->gc_flag  = IMPREC_NOT_COPIED;

  regist_imp_entry(new_exref);

  return new_exref->to_exref;
}

extern struct generator_object_method_table*
get_exref_methtab()
{
  return &G_method_table;
}

#endif /* DIST */
